/*  FileApart splits files into smaller fragments and reassembles these 
 *  fragments at the desired destination.
 *  Copyright (C) 2008 by Markus Walther
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package fileapart;

import java.io.*;


/**
 *
 * @author Markus Walther (mwthp@web.de)
 */
public class chopAndMergeFiles {
  //Mandatory input for the File Chopping process
  String sFileToChop;  //Source file
  String sFragDestDir; //Folder which is to contain the generated fragments 
  long FragSize;      //Fragment size in Bytes

  //Mandatory input for the file reassembly process
  String sFirstFragment;   //First fragment within the fragment source folder
  String sFileDestDir; //Destination folder for the reassembled file

  //attributes regarding the chopping of a file

  //source file
  File sFile; //File to chop, e.g. the source file
  long sFileLength;  //length of the source file
  FileInputStream sFileInputStream;  //File to chop as FileInputStream
  long sPos;
  //Folder which is to contain the generated fragments 
  File FragDestDirFile;
  long nFrags;  //number of fragments to be written
  long maxFrags; //maximal number of fragment files to be written
  long nBufferFillings; //how often will the buffer be used in order to write a fragment?
  String curFragFileName; //current fragment file name
  long lFragBytes;  //number of bytes contained within the fragment files
  
  //attributes regarding the reassembly of a file
  File[] fragFileList;  //list of fragment Files
  String sOrgFileName;  //name of the original file
  File fragSourceDir;
  long reassemblyBytesWritten;
  
  //used with both operations
  //buffer
  int MaxBufSize; //will be set to a default value within the constructor
  int BufSize;    //will be calculated
  byte buffer[];  //the read/write buffer
  //format string
  String appFormat;
  boolean overwriteFiles;
  /** A return status code - No */
  public static final int RET_NO = 0;
  /** A return status code - Yes */
  public static final int RET_YES = 1;
  /** A return status code - All */
  public static final int RET_ALL = 2;
  /** Copyright information */
  public static final String CHOP_AND_MERGE_FILES_COPYRIGHT="FileApart Version 1.0\n\nCopyright (c) 2008 by Markus Walther\n"+
    "This program comes with ABSOLUTELY NO WARRANTY;\n" +
    "This is free software, and you are welcome to \nredistribute it under certain conditions; \n\n" +
    "See http://www.gnu.org/licenses/gpl-3.0.html for details.";


  /**
   * Sets some internal values, like the size of the internal buffer, and displays the version 
   * and copyright information.
   */
  public chopAndMergeFiles() {
    MaxBufSize = 1024 * 1024;   //1 MB
    appFormat = ".part.%03d";
    maxFrags = 1000;
    overwriteFiles = false;
    System.out.println(CHOP_AND_MERGE_FILES_COPYRIGHT + "\n");
  }
  /**
   * Calls System.exit(0).
   */
  public void onExit() {
    System.exit(0);
  }

  //file chopping related methods
  /**
   * Sets the absolute path of the file to be chopped into fragments.
   * @param FileToChop absolute path of the file to be chopped.
   */
  public void setFileToChop(String FileToChop) {
    this.sFileToChop = FileToChop;
  }
  /**
   * Sets the folder which is to contain the generated fragment files.
   * @param FragDestDir absolute path to the desired folder.
   */
  public void setFragDestDir(String FragDestDir) {
    this.sFragDestDir = FragDestDir;
  }

  /**
   * Sets the maximal fragment file size in bytes.
   * @param FragSize maximal fragment file size in bytes.
   */
  public void setFragSize(long FragSize) {
    this.FragSize = FragSize;
  }
  
  /**
   * Checks whether all mandatory input variables for the chopping of a file are set.
   * E.g. the file to chop, the folder for holding the generated fragments 
   * and the maximal size of a fragment file. This routine does not check any validity
   * of the given variables, but only proofs their existence.
   * @return true, if the variables are set, false otherwise.
   */
  public boolean checkMandatoriesForChopping() {
    boolean res = true;
    if (isVoidString(sFileToChop)) {
      res = false;
    }
    if (isVoidString(sFragDestDir)) {
      res = false;
    }
    if (FragSize == 0) {
      res = false;
    }
    return res;

  }
  /**
   * Performs inital tests and calculations.
   * Use this method after having collected all mandatory data from the GUI.
   * Called by chopFile.
   * @return true, if the necessary input variables are correctly set, otherwise false.
   * */
  public boolean initFileChopping() {
    if(!checkMandatoriesForChopping()){
      showMessage("Not all necessary input variables for file chopping are set.");
      return false;
    }
    sFile = new File(sFileToChop);
    if (!sFile.isFile()) {
      showMessage(sFileToChop + " is not a file!");
      return false;
    }
    FragDestDirFile = new File(sFragDestDir);
    //should be a folder.
    if (!FragDestDirFile.isDirectory()) {
      showMessage(sFragDestDir + " is not a folder!");
      return false;
    }

    //no valid FragSize, should not happen at this point
    if (FragSize <= 0) {
      showMessage("Wrong fragment size given!");
      return false;
    }
    //get the size of the source file
    try {
      sFileLength = sFile.length();
    } catch (SecurityException se) {
      showMessage(se.getMessage());
      return false;
    }
    if(sFileLength == 0)
    {
      showMessage(sFileToChop + "has zero bytes!");
      return false;
    }
    if (FragSize > sFileLength) {
      FragSize = sFileLength;
    }  //would be stupid, otherwise.
    //if a fragment is smaller than the default size of the buffer, 
    //the size of the buffer will be set accordingly
    BufSize = (FragSize < MaxBufSize) ? (int) FragSize : MaxBufSize;
    buffer = new byte[BufSize];
    //how many fragment files are to be written?
    nFrags = sFileLength / FragSize;
    //will the last fragment file be smaller than the others?
    if ((sFileLength % FragSize) > 0) {
      nFrags++;
    }
    //how often needs the buffer to be refilled in order 
    //to write a regular fragment file?
    nBufferFillings = FragSize / BufSize;
    if ((FragSize % BufSize) > 0) {
      nBufferFillings++;
    }
    if(nFrags>maxFrags){
      showMessage("Maximal number of fragment files exceeded (which would be as much as " + maxFrags + ")!");
      return false;
    }
    //open sFile as FileInputStream in order to read from it
    try {
      sFileInputStream = new FileInputStream(sFile);
    } catch (FileNotFoundException fnf) {
      showMessage(fnf.getMessage());
      return false;
    } catch (SecurityException se) {
      showMessage(se.getMessage());
      return false;
    }

    return true;
  }
  /**
   * Checks whether the current channel position can be accessed.
   * @return the boolean result.
   */
  protected boolean getPos() {
    try {
      sPos = sFileInputStream.getChannel().position();
    } catch (IOException ioe) {
      showMessage(ioe.getMessage());
      return false;
    }
    return true;
  }
  /**
   * Displays the current state of the some internal parameters regarding the process of 
   * creating fragment files.
   */
  public void chopTestOutput() {
    showMessage("\nFILE CHOPPING\n");
    if (!initFileChopping()) {
      return;
    }
    showMessage("Attribute check:");
    showMessage("File to chop: " + sFile.getAbsolutePath());
    showMessage("Target directory " +
            FragDestDirFile.getAbsolutePath());
    showMessage("Size of source file: " + sFileLength + " Bytes");
    showMessage("Fragment size: " + FragSize + " Bytes.");
    showMessage("Number of fragments: " + nFrags);
    /*
    showMessage("Buffer size: " + BufSize + " Bytes");
    showMessage("Maximum of buffer fillings per fragment: " + nBufferFillings);
    if (getPos()) {
      showMessage("Current channel position: " + sPos);
    }
    */
    if (createFragFilename(0)) {
      showMessage("First fragment will be named: " + curFragFileName);
    }
    showMessage("");
  }
  /**
   * Example: FileToChop == "hello.img", FragDestDir == "/home/user/frag".
   * Then the first fragment file will be named "/home/user/frag/hello.img.part000".
   * @return true if the name creation succeeded, otherwise false.
   */
  protected boolean createFragFilename(int i) {
    String resstr;

    resstr = FragDestDirFile.getAbsolutePath() + File.separator +
            sFile.getName() + String.format(appFormat, i);
    curFragFileName = resstr;
    return true;
  }

  /**
   * Writes a single fragment file.
   * @param i number of the fragment file.
   * @return true, if the file could be written, otherwise false.
   */
  protected boolean writeFragFile(int i) {
    FileOutputStream fragFile;
    long j;
    long baw = 0; //bytes already written
    long btw; //remaining bytes to be written
    int br;

    if (!createFragFilename(i)) {
      return false;
    }
    if(!checkFileOverwrite(curFragFileName)){
      return false;
    }
    try {  //open a new Fragment file for output
      fragFile = new FileOutputStream(curFragFileName);
    } catch (FileNotFoundException fnf) {
      showMessage(fnf.getMessage());
      return false;
    } catch (SecurityException se) {
      showMessage(se.getMessage());
      return false;
    }
    for (j = 0; j < nBufferFillings; j++) {
      //if(!getPos())
      //  return false;
      btw = FragSize - baw;
      btw = (btw > BufSize) ? BufSize : btw; //making sure that not more than BufSize bytes are written
      try {  //read btw bytes from the source stream
        br = sFileInputStream.read(buffer, 0, (int) btw);
      } catch (IOException ioe) {
        showMessage(ioe.getMessage());
        return false;
      }
      if (br <= 0) {
        break;
      }  //finished, no more source bytes to process

      try {  //write the bytes, which has been just read from the source to the fragment file
        fragFile.write(buffer, 0, br);
      } catch (IOException ioe) {
        showMessage(ioe.getMessage());
        return false;
      }
      baw += br;
      lFragBytes += br;
    }

    showMessage("Successfully wrote " + curFragFileName + ".");
    try {
      fragFile.close();
    } catch (IOException ioe) {
      showMessage(ioe.getMessage());
      return false;
    }
    return true;
  }
  /**
   * chopFile tries to write all fragment files into the desired fragment file folder.
   * It also checks whether the neccessary inputs are sufficient
   * @param verboseOutput If you set this to true, you'll get inormation about
   * some internal settings.
   * @return true, if the file has successfully been chopped into fragments, otherwise false.
   */ 
  public boolean chopFile(boolean verboseOutput) {
    int i;

    if (!initFileChopping()) {
      showMessage("initFileChopping failed.");
      if(verboseOutput){
        chopTestOutput();
      }
      return false;
    }
    if(verboseOutput)
      chopTestOutput();
    
    lFragBytes = 0;
    overwriteFiles = false;
    for (i = 0; i < nFrags; i++) {
      if (!writeFragFile(i)) {
        return false;
      }
    }
    if(sFileLength == lFragBytes)
      showMessage("Success!");
    else
      showMessage("Failure.");
    return true;
  }

  /**
   * Sets the folder in which the reassembled (the original) file will be stored.
   * @param FileDestDir The folder in which the reassembled (the original) file will be stored.
   */
  public void setFileDestDir(String FileDestDir) {
    this.sFileDestDir = FileDestDir;
  }

  /**
   * Sets the location of the first fragment file
   * @param FirstFragment absolute path to the first fragment file
   */
  public void setFirstFragment(String FirstFragment) {
    this.sFirstFragment = FirstFragment;
  }

  /**
   * Checks if all necessary strings are set.
   * E.g. the first fragment file name and 
   * the name of the destination folder.
   * @return true if so, otherwise false.
   */
  public boolean checkMandatoriesForAssembling() {
    boolean res=true;
    if (isVoidString(sFileDestDir)) {
      res = false;
    }
    if (isVoidString(sFirstFragment)) {
      res = false;
    }
    else 
    {
      if(!isFirstFragFile(sFirstFragment)){
        res = false;
      }
    }
    return res;
  }

  /**
   * Checks whether the file name ends with "part.000".
   * @param fileName the file to check
   * @return true, if it matches the criteria, otherwise false.
   */
  protected boolean isFirstFragFile(String fileName) {
    return (fileName.endsWith(String.format(appFormat, 0)));
  }

  /**
   * Extracts the original file name from the name of the first fragment file.
   * @param firstFragFile the name of the first fragment file
   * @return true if the file name could be extracted, otherwise false.
   */
  protected boolean setOrgFileName(File firstFragFile) {  //strip off ".partXXX".
    String fname;
    int idx;

    fname = firstFragFile.getName();
    idx = fname.indexOf(String.format(appFormat, 0));
    if (idx > 1) {
      sOrgFileName = fname.substring(0, idx);
      return true;
    }
    return false;
  }

  /**
   * Creates a list of fragment files.
   * Call this method twice, first to get the count of matching files,
   * and a second time to fill the list.
   * @param fileList file list.
   * @param create false: get the number of matching files, true: fill the list
   * @return true if the list could be created, otherwise false.
   */

  protected int checkOrFillList(File[] fileList, boolean create){ 
    boolean found = true;
    int i, idx = 0;
    String nextFrag;

    while (found) {
      found = false;
      for (i = 0; i < fileList.length; i++) {
        nextFrag = sOrgFileName + String.format(appFormat, idx);
        if (fileList[i].getName().equals(nextFrag)) {
          if(create)
            fragFileList[idx] = fileList[i].getAbsoluteFile();  //create an ordered list
          idx++;
          found = true;
        }
      }
    }
    return idx;
  }
  /**
   * creates the internal fragment file list.
   * @return true if successful, otherwise false.
   */
  protected boolean createFragList() {
    File[] fileList;
    int count;

    fileList = fragSourceDir.listFiles();
    if (fileList == null) {
      return false;
    }
    count = checkOrFillList(fileList, false);
    if(count == 0)
      return false;
    fragFileList = new File[count];
    return checkOrFillList(fileList, true) > 0;
  }

  /**
   * Performs inital tests and calculations.
   * Use this method after having collected all mandatory data from the GUI.
   * Called by reassembleFile.
   * @return true, if the necessary input variables are correctly set, otherwise false.
   * */
  public boolean initFileAssembly() {
    File curFragFile;
    File FileDestDir;
  
    
    if(!checkMandatoriesForAssembling()){
      showMessage("Not all necessary input variables for reassembling are set correctly.");
      return false;
    }

    curFragFile = new File(sFirstFragment);
    if (!curFragFile.isFile()) {
      showMessage(curFragFile + " is not a file!");
      return false;
    }
    fragSourceDir = curFragFile.getParentFile();
    if (fragSourceDir == null) {
      showMessage("Could not obtain the parent directory of " + sFirstFragment);
    }
    FileDestDir = new File(sFileDestDir);
    //should be a folder.
    if (!FileDestDir.isDirectory()) {
      showMessage(sFileDestDir + " is not a folder!");
      return false;
    }
    if(!setOrgFileName(curFragFile))
      return false;
    if(true)
    {
      if(!createFragList())
        return false;
    }
    BufSize = (MaxBufSize<curFragFile.length())? MaxBufSize : (int)curFragFile.length();
    nBufferFillings = curFragFile.length() / BufSize;
    if((curFragFile.length() % BufSize) > 0)
      nBufferFillings++;
    buffer = new byte[BufSize];
    return true;
  }
  
  /**
   * Reassembles the original file from the given fragment files.
   * @param verboseOutput set this parameter to true if you want to get additional information.
   * @return true if the reassembly succeeded, otherwise false.
   */
  public boolean reassembleFile(boolean verboseOutput){
    FileInputStream curFragFileIS;
    FileOutputStream assembledFile;
    int rb,rdsum;
    long curSize;
    String TargetFileName;
    
    if(!initFileAssembly()){
      showMessage("initFileAssembly failed!");
      if(verboseOutput){
        mergeTestOutput();
      }
      return false;
    }
    if(verboseOutput){
      mergeTestOutput();
    }
    reassemblyBytesWritten = 0;
    TargetFileName = sFileDestDir + File.separator + sOrgFileName;
    overwriteFiles = false;
    if(!checkFileOverwrite(TargetFileName))
      return false;
    
    try{
      assembledFile = new FileOutputStream(TargetFileName);
    } catch(FileNotFoundException fnf){
      showMessage(fnf.getMessage());
      return false;
    } catch (SecurityException se){
      showMessage(se.getMessage());
      return false;
    }
    
    //starting the reassembly
    for(int i=0; i<fragFileList.length; i++){
      try{
        curFragFileIS = new FileInputStream(fragFileList[i]);
      } catch (FileNotFoundException fnf) {
        showMessage(fnf.getMessage());
        return false;
      }
      curSize = fragFileList[i].length();
      rdsum = 0;
      for(int j=0; j<nBufferFillings;j++){
        try{
          rb=curFragFileIS.read(buffer);
        } catch (IOException ioe) {
          showMessage(ioe.getMessage());
          return false;
        }
        rdsum += rb;
        //now do the writing
        try{
          assembledFile.write(buffer, 0, rb);
        } catch(IOException ioe) {
          showMessage(ioe.getMessage());
          return false;
        }
        reassemblyBytesWritten += rb;
        if(rdsum == curSize)  //no more bytes to read
        {
          try{
            curFragFileIS.close();
          } catch(IOException ioe) {
            showMessage(ioe.getMessage());
            return false;
          }
          break;
        }
      }
      showMessage("Fragment "+i+" added to "+TargetFileName+".");
     }
    try{
      assembledFile.close();
    } catch(IOException ioe) {
      showMessage(ioe.getMessage());
      return false;
    }
    showMessage("Successfully wrote "+reassemblyBytesWritten+" Bytes to " +sOrgFileName);
    return true;
  }
  
  public void mergeTestOutput(){
    showMessage("\nFILE ASSEMBLY\n");
    if(!initFileAssembly())
    {
      return;
    }
    showMessage("Name of the target file: " + sFileDestDir + File.separator + sOrgFileName);
    showMessage("Fragment file folder: "+ fragSourceDir.getAbsolutePath());
    showMessage("");
    if(fragFileList!=null){
      showMessage("Found "+ fragFileList.length + " Fragment files, which read out to:");
      for(int i=0; i<fragFileList.length;i++){
        showMessage("Fragment file #" + i + ": " + fragFileList[i].getName());
      }
    }
    showMessage("\nBuffer size: " + BufSize + " bytes.");
    showMessage("Maximum of buffer fillings per fragment file: " + nBufferFillings);
  }
  
  /**
   * Displays a message to the console.
   * @param message the message to be displayed.
   */
  public void showMessage(String message){
    System.out.println(message);
  }
  
  /**
   * tries to determine the language of the operating system.
   * @return a string which contains information about the language.
   */
  public String getOSLanguage(){
    return System.getenv("LANG");
  }
  
  /**
   * Checks whether s==null or s is empty.
   * @param s string to check.
   * @return false if s is not null and not empty, otherwise true.
   */
  protected boolean isVoidString(String s){
    if (s==null)
      return true;
    if (s.isEmpty())
      return true;
    return false;
  }
  /**
   * checks whether a "overwrite file?" dialog is to be launched.
   * @param filename
   * @return true if the file should be overwritten, otherwise false.
   */
  protected boolean checkFileOverwrite(String filename){
    int overwrite=0;
    File eFile = new File(filename);
    if (!eFile.isFile())
      return true;  //nothing to overwrite there
    if(overwriteFiles)
      return true;

    overwrite = launchOverwriteDialog(filename);
    
    if(overwrite == RET_ALL)
      overwriteFiles = true;
 
    if(overwrite > 0)
      return true;
    
    return false;
  }
  public int launchOverwriteDialog(String fname){
    showMessage("File "+fname+"already exists!");
    showMessage("Overwrite (y/a/n)?");
    int result;
    try{
      result = System.in.read();
    } catch (IOException ioe){
      showMessage(ioe.getMessage());
      return RET_NO;
    }
    switch(result){
      case 'y': return RET_YES;
      case 'a': return RET_ALL;
      case 'n': return RET_NO;
      default: return RET_NO;
    }
  }
}
